//
//  $Id: i2c.c 42 2008-10-04 18:40:36Z jcw $
//  $Revision: 42 $
//  $Author: jcw $
//  $Date: 2008-10-04 14:40:36 -0400 (Sat, 04 Oct 2008) $
//  $HeadURL: http://tinymicros.com/svn_public/arm/lpc2148_demo/trunk/i2c/i2c.c $
//

#include <inttypes.h>
#include <stdbool.h>

#include "FreeRTOS.h"
#include "task.h"

#include "LPC17xx.h"
#include "i2c.h"


/* Register bit definitions */
#define I2C_CONSET_AA               (1 << 2)
#define I2C_CONSET_SI               (1 << 3)
#define I2C_CONSET_STO              (1 << 4)
#define I2C_CONSET_STA              (1 << 5)
#define I2C_CONSET_I2EN             (1 << 6)

#define I2C_CONCLR_AAC              (1 << 2)
#define I2C_CONCLR_SIC              (1 << 3)
#define I2C_CONCLR_STAC             (1 << 5)
#define I2C_CONCLR_I2ENC            (1 << 6)
#define I2C_CONCLR_MASK             (0x0000006C)

#define I2C_STAT_STATMASK           (0x1F << 3)



//
//  Default timeout, in milliseconds for generic read/write
//
#define I2C_DEFAULT_TIMEOUT 100
#undef  I2C_DEBUG

//
//
//
typedef enum
{
  ACKPOLL_NO = 0,
  ACKPOLL_YES
}
ackPoll_e;

//
//
//
static volatile int i2cBusInUse;
static uint8_t i2cAddress;
static uint8_t *i2cDataBuffer;
static int i2cDataLenWrite;
static int i2cDataLenRead;
static unsigned int i2cTimeoutInTicks = I2C_DEFAULT_TIMEOUT / portTICK_RATE_MS;
static int i2cDataCnt;
static ackPoll_e i2cAckPoll = ACKPOLL_NO;

i2cErr_e i2cErrno = I2CERR_NONE;


void I2C1_IRQHandler (void);






void I2C1_IRQHandler (void)
{
  i2cErrno = (LPC_I2C1->I2STAT & I2C_STAT_STATMASK);

  switch (i2cErrno)
  {
    //
    //  Transmit conditions
    //
    case 0x00:
      {
        LPC_I2C1->I2CONSET = I2C_CONSET_STO | I2C_CONSET_AA;
        i2cAckPoll = ACKPOLL_NO;
        i2cBusInUse = false;
      }
      break;

    case 0x08:
    case 0x10:
      {
        LPC_I2C1->I2DAT = i2cAddress;
      }
      break;

    case 0x18:
    case 0x28:
      {
        i2cAckPoll = ACKPOLL_NO;

        if (i2cDataLenWrite && (i2cDataCnt < i2cDataLenWrite))
        {
          LPC_I2C1->I2DAT = i2cDataBuffer [i2cDataCnt++];
          LPC_I2C1->I2CONCLR = I2C_CONCLR_STAC;
        }
        else
        {
          if (!i2cDataLenRead)
          {
            LPC_I2C1->I2CONCLR = I2C_CONCLR_STAC;
            LPC_I2C1->I2CONSET = I2C_CONSET_STO;
            i2cErrno = I2CERR_NONE;
            i2cBusInUse = false;
          }
          else
          {
            i2cAddress |= 0x01;
            i2cDataCnt = 0;
            LPC_I2C1->I2CONSET = I2C_CONSET_STA;
          }
        }
      }
      break;

    case 0x58:
      {
        i2cDataBuffer [i2cDataCnt++] = LPC_I2C1->I2DAT;
        i2cErrno = I2CERR_NONE;
      }

    case 0x20:
    case 0x48:
      {
        if (i2cAckPoll)
          break;
      }

    case 0x30:
      {
        LPC_I2C1->I2CONCLR = I2C_CONCLR_STAC;
        LPC_I2C1->I2CONSET = I2C_CONSET_STO;
        i2cBusInUse = false;
      }
      break;

    case 0x38:
      {
        LPC_I2C1->I2CONSET = I2C_CONSET_STA;
      }
      break;

    //
    //  Receive byte conditions (fall through is intentional)
    //
    case 0x50:
      {
        i2cDataBuffer [i2cDataCnt++] = LPC_I2C1->I2DAT;
      }

    case 0x40:
      {
        if (i2cDataCnt < i2cDataLenRead - 1)
        {
          LPC_I2C1->I2CONCLR = I2C_CONCLR_STAC;
          LPC_I2C1->I2CONSET = I2C_CONSET_AA;
        }
        else
          LPC_I2C1->I2CONCLR = I2C_CONCLR_STAC | I2C_CONCLR_AAC;
      }
      break;

    case I2CERR_NOINFO :
      break;

    default:
      {
        LPC_I2C1->I2CONCLR = I2C_CONCLR_I2ENC;
        i2cAckPoll = ACKPOLL_NO;
        i2cBusInUse = false;
      }
      break;
  }

  LPC_I2C1->I2CONCLR = I2C_CONSET_SI;
}


//
// i2c1Init
//
void i2cInit (void)
{
  i2cBusInUse = false;
  i2cAckPoll = ACKPOLL_NO;


  LPC_I2C1->I2CONCLR = I2C_CONCLR_MASK;
  LPC_I2C1->I2CONSET = I2C_CONSET_I2EN;
  LPC_I2C1->I2SCLL   = 100;
  LPC_I2C1->I2SCLH   = 100;

  // Enable interrupt
  NVIC_EnableIRQ( I2C1_IRQn );
}

//
//  Waits for the interrupt handler to mark the I2C bus as no longer in use, then
//  return 0 on success or -1 if any error occurred.
//
static int i2cWaitComplete (int ticks)
{
  if (i2cBusInUse)
  {
    portTickType theFuture;

    for (theFuture = xTaskGetTickCount() + ticks; i2cBusInUse; )
    {
      if (xTaskGetTickCount() > theFuture)
      {
        LPC_I2C1->I2CONCLR = I2C_CONCLR_I2ENC;
        i2cErrno = I2CERR_TIMEOUTWC;
        return -1;
      }
    }
  }

  return (i2cErrno == I2CERR_NONE) ? 0 : -1;
}

//
//  Disables interrupts then starts sending repeated STARTs until the slave device
//  returns ACK.  This indicates the slave has completed any operation in progress.
//
static int i2cPoll (uint8_t address)
{
  portTickType theFuture = xTaskGetTickCount() + 10;

  NVIC_DisableIRQ( I2C1_IRQn );

  while (xTaskGetTickCount() < theFuture)
  {
    LPC_I2C1->I2CONCLR = I2C_CONSET_SI;
    LPC_I2C1->I2CONSET = I2C_CONSET_STA;

    while (!(LPC_I2C1->I2CONSET & I2C_CONSET_SI))
      ;

    LPC_I2C1->I2CONCLR = I2C_CONCLR_STAC | I2C_CONCLR_SIC;
    LPC_I2C1->I2DAT = address & ~0x01;

    while (!(LPC_I2C1->I2CONSET & I2C_CONSET_SI))
      ;

    if ((i2cErrno = LPC_I2C1->I2STAT) == I2CERR_SLAWTX_ACKRX)
      break;
  }

  if (i2cErrno != I2CERR_SLAWTX_ACKRX)
    i2cErrno = I2CERR_TIMEOUTACKPOLL;

  LPC_I2C1->I2CONCLR = I2C_CONCLR_SIC;
  LPC_I2C1->I2CONSET = I2C_CONSET_STO | I2C_CONSET_AA;

  while (LPC_I2C1->I2CONSET & I2C_CONSET_STO)
    ;

  NVIC_EnableIRQ( I2C1_IRQn );

  return (i2cErrno == I2CERR_SLAWTX_ACKRX) ? 0 : -1;
}

//
//
//
static int i2cKickOff (uint8_t address, uint8_t *buffer, int bufferLenWrite, int bufferLenRead, ackPoll_e ackPoll)
{
  //
  //  Determine if our first operation will be a write or read.  If both, the
  //  write always occurs first.
  //
  if (bufferLenWrite)
    address &= ~0x01;
  else if (bufferLenRead)
    address |= 0x01;
  else
    address &= ~0x01;

  //
  //  Wait until last I2C operation has finished.  
  //
  if (i2cBusInUse && i2cWaitComplete (i2cTimeoutInTicks))
  {
    i2cErrno = I2CERR_TIMEOUT;
    return -1;
  }

  //
  //  Mark bus as in use, save the address, buffer and length
  //
  i2cBusInUse = true;
  i2cAddress = address;
  i2cDataBuffer = buffer;
  i2cDataLenWrite = bufferLenWrite;
  i2cDataLenRead = bufferLenRead;
  i2cDataCnt = 0;
  i2cAckPoll = ackPoll;

  //
  //  Clear any outstanding bits, enable the interface, and send a start bit.
  //  Once the start bit goes out, the interrupt drive state machine begins.
  //
  LPC_I2C1->I2CONCLR = I2C_CONCLR_MASK;
  LPC_I2C1->I2CONSET = I2C_CONSET_I2EN;
  LPC_I2C1->I2CONSET = I2C_CONSET_STA;

  return 0;
}

//
//
//
static int i2cWriteBufferEx (uint8_t address, uint8_t *buffer, uint32_t bufferLength, int milliseconds)
{
  int r;

  if (!(r = i2cKickOff (address, buffer, bufferLength, 0, ACKPOLL_NO)))
    r = i2cWaitComplete (milliseconds / portTICK_RATE_MS);

  return r;
}

static int i2cReadBufferEx (uint8_t address, uint8_t *buffer, uint32_t bufferLength, int milliseconds)
{
  int r;

  if (!(r = i2cKickOff (address, buffer, 0, bufferLength, ACKPOLL_NO)))
    r = i2cWaitComplete (milliseconds / portTICK_RATE_MS);

  return r;
}

static int i2cWriteReadBufferEx (uint8_t address, uint8_t *buffer, uint32_t putLength, uint32_t getLength, int milliseconds)
{
  int r;

  if (!(r = i2cKickOff (address, buffer, putLength, getLength, ACKPOLL_NO)))
    r = i2cWaitComplete (milliseconds / portTICK_RATE_MS);

  return r;
}

//
//
//
void i2cSetTimeout (unsigned int timeoutInMilliseconds)
{
  if (timeoutInMilliseconds < portTICK_RATE_MS)
    timeoutInMilliseconds = portTICK_RATE_MS;

  i2cTimeoutInTicks = timeoutInMilliseconds / portTICK_RATE_MS;
}


//
//  DANGER, WILL ROBINSON!  The callers buffer must persist until we're done
//
int i2cWriteBuffer (uint8_t address, uint8_t *buffer, uint32_t bufferLength)
{
  return i2cWriteBufferEx (address, buffer, bufferLength, i2cTimeoutInTicks);
}

int i2cReadBuffer (uint8_t address, uint8_t *buffer, uint32_t bufferLength)
{
  return i2cReadBufferEx (address, buffer, bufferLength, i2cTimeoutInTicks);
}

int i2cWriteReadBuffer (uint8_t address, uint8_t *buffer, uint32_t putLength, uint32_t getLength)
{
  return i2cWriteReadBufferEx (address, buffer, putLength, getLength, i2cTimeoutInTicks);
}

int i2cWriteBufferPoll (uint8_t address, uint8_t *buffer, uint32_t bufferLength)
{
  int r;

  if (!(r = i2cKickOff (address, buffer, bufferLength, 0, ACKPOLL_NO)))
    if (!(r = i2cWaitComplete (i2cTimeoutInTicks)))
      r = i2cPoll (address);

  return r;
}

int i2cWriteReadBufferPoll (uint8_t address, uint8_t *buffer, uint32_t putLength, uint32_t getLength)
{
  int r;

  if (!(r = i2cKickOff (address, buffer, putLength, getLength, ACKPOLL_NO)))
    if (!(r = i2cWaitComplete (i2cTimeoutInTicks)))
      r = i2cPoll (address);

  return r;
}
