
#include <stdint.h>

#include "LPC17xx.h"
#include "init.h"


/* PLL frequency is f_xtal * 2 * (M / N). Here: 399.0 MHz */
#define PLL0_M            (133)
#define PLL0_N            (8)
#define PLL0CFG_Val       (((PLL0_N - 1) << 16) | (PLL0_M - 1))

/* USB PLL configuration. P=2 --> Fvco=4*Fusb=192 MHz, M=4 --> Fusb=4*Fxtal */
#define PLL1_P            (2)
#define PLL1_M            (4)
#define PLL1CFG_Val       ((PLL1_P << 4) | (PLL1_M - 1))

/* Clock divider (f_cco / f_cclk). Must be an even number */
#define CLOCK_DIV         (4)

/* Number of cycles per flash access */
#define FLASH_WAITSTATES  (5)





/**
 * Initialize the non-peripheral related part of the system (PLL, pins)
 *
 * @param  none
 * @return none
 *
 * @brief  Setup the system.
 */
void cpuInit( void )
{
    /*----- Peripheral clocks (default: 1/4) -----------------------------*/
    LPC_SC->PCLKSEL0 = 0x00000003;      /* WDT=1/8                                    */
    LPC_SC->PCLKSEL1 = 0x00800000;      /* I2S=1/2                                    */
    LPC_SC->PCONP    = 0xA8E98406;      /* TIMER0(1), TIMER1(2), SSP1(10),            */
                                        /* GPIO(15), RIT(16), I2C1(19), SSP0(21)      */
                                        /* TIMER2(22), TIMER3(23), I2S(27), DMA(29)   */
                                        /* USB(31)                                    */

#if BOARD_TYPE == BOARD_ROLF1
    /*----- Pin connections ----------------------------------------------*/
    LPC_PINCON->PINSEL0 = 0x0009500F;   /* P0.0=SDA1          P0.1=SCL1           */
                                        /* P0.6=I2SRX_SDA     P0.7=I2STX_CLK      */
                                        /* P0.8=I2STX_WS      P0.9=MOSI1          */
    LPC_PINCON->PINSEL1 = 0x14200000;   /* P0.26=AOUT                             */
                                        /*                    P0.29=USB_D+        */
                                        /* P0.30=USB_D-                           */
    LPC_PINCON->PINSEL3 = 0xA0000000;   /* P1.30=VBUS         P1.31=SCK1          */
    LPC_PINCON->PINSEL4 = 0x00040000;   /*                    P2.9=USB_CONNECT    */
    LPC_PINCON->PINSEL9 = 0x04000000;   /*                    P4.29=TX_MCLK       */

    /*----- Internal pull-up and pull-down -------------------------------*/
    LPC_PINCON->PINMODE0 = 0x008A900A;  /* P0.0=---           P0.1=---            */
                                        /* P0.6=Repeater      P0.7=---            */
                                        /* P0.8=---           P0.9=---            */
                                        /*                    P0.11=---           */
    LPC_PINCON->PINMODE1 = 0x28280000;  /*                    P0.25=---           */
                                        /* P0.26=---                              */
                                        /*                    P0.29=---           */
                                        /* P0.30=---                              */
    LPC_PINCON->PINMODE2 = 0x0000020A;  /* P1.0=---           P1.1=---            */
                                        /* P1.4=---                               */
    LPC_PINCON->PINMODE3 = 0xA0000020;  /* P1.18=---                              */
                                        /* P1.30=---          P1.31=---           */
    LPC_PINCON->PINMODE4 = 0x0008AA00;  /* P2.4=---           P2.5=---            */
                                        /* P2.6=---           P2.7=---            */
                                        /* P2.9=---                               */
    LPC_PINCON->PINMODE9 = 0x0A000000;  /* P4.28=---          P4.29=---           */

    /* ---- GPIO's -------------------------------------------------------*/
    LPC_GPIO0->FIODIR = 0x02000800;     /* P0.11=SI570_OE, P0.25=GUI_LICHT        */
    LPC_GPIO0->FIOCLR = (1 << 25);
    LPC_GPIO1->FIODIR = 0x00040013;     /* P1.0=GUI_SSEL, P1.1=GUI_RESET          */
                                        /* P1.4=GUI_RESET                         */
                                        /* P1.18=LED                              */
    LPC_GPIO2->FIODIR = 0x000000F0;     /* P2.4=TEILER1, P2.5=TEILER2             */
                                        /* P2.6=ENAREG1, P2.7=ENAREG2             */
    LPC_GPIO0->FIOSET = (1 << 11);      /* Enable Si570                           */
    LPC_GPIO2->FIOSET = (1 << 6) | (1 << 7);
                                        /* Enable regulators                      */
    LPC_GPIO4->FIODIR &= ~0x10000000;   /* P4.28=UDA1361_PWON (floating = 0 dB)   */
#endif

#if BOARD_TYPE == BOARD_KAI1
    /*----- Pin connections ----------------------------------------------*/
    LPC_PINCON->PINSEL0 = 0x0001500F;   /* P0.0=SDA1          P0.1=SCL1           */
                                        /* P0.6=I2SRX_SDA     P0.7=I2STX_CLK      */
                                        /* P0.8=I2STX_WS                          */
    LPC_PINCON->PINSEL1 = 0x14000000;   /*                    P0.29=USB_D+        */
                                        /* P0.30=USB_D-                           */
    LPC_PINCON->PINSEL3 = 0x20000000;   /* P1.30=VBUS                             */
    LPC_PINCON->PINSEL4 = 0x00040000;   /*                    P2.9=USB_CONNECT    */
    LPC_PINCON->PINSEL9 = 0x04000000;   /*                    P4.29=TX_MCLK       */

    /*----- Internal pull-up and pull-down -------------------------------*/
    LPC_PINCON->PINMODE0 = 0x0002900A;  /* P0.0=---           P0.1=---            */
                                        /* P0.6=Repeater      P0.7=---            */
                                        /* P0.8=---                               */
    LPC_PINCON->PINMODE1 = 0x28002000;  /* P0.22=---                              */
                                        /*                    P0.29=---           */
                                        /* P0.30=---                              */
    LPC_PINCON->PINMODE2 = 0x002A0200;  /* P1.4=---                               */
                                        /* P1.8=---           P1.9=---            */
                                        /* P1.10=---                              */
    LPC_PINCON->PINMODE3 = 0x200000A0;  /* P1.18=---          P1.19=---           */
                                        /* P1.30=---                              */
    LPC_PINCON->PINMODE4 = 0x000AA800;  /*                    P2.5=---            */
                                        /* P2.6=---           P2.7=---            */
                                        /* P2.8=---           P2.9=---            */
    LPC_PINCON->PINMODE9 = 0x0A000000;  /* P4.28=---          P4.29=---           */

    /* ---- GPIO's -------------------------------------------------------*/
    LPC_GPIO0->FIODIR = (1 << 22);      /* P0.22=ENAREG1                          */
    LPC_GPIO0->FIOSET = (1 << 22);      /* Enable regulator1                      */
    LPC_GPIO1->FIODIR = 0x000C0710;     /* P1.4=IQSWAP                            */
                                        /* P1.8=TEILER2       P1.9=TEILER1        */
                                        /* P1.10=SI570_OE                         */
                                        /* P1.18=LED          P1.19=ENAREG2       */
    LPC_GPIO1->FIOCLR = (1 << 19);      /* Enable regulator2                      */
    LPC_GPIO1->FIOSET = (1 << 10);      /* Enable Si570                           */
    LPC_GPIO2->FIODIR = 0x000001E0;     /*                    P2.5=PRESELECT1     */
                                        /* P2.6=PRESELECT2    P2.7=PRESELECT3     */
                                        /* P2.8=PRESELECT4                        */
    LPC_GPIO4->FIODIR &= ~0x10000000;   /* P4.28=UDA1361_PWON (floating = 0 dB)   */
#endif

#if BOARD_TYPE == BOARD_KAI2
    /*----- Pin connections ----------------------------------------------*/
    LPC_PINCON->PINSEL0 = 0x0001500F;   /* P0.0=SDA1          P0.1=SCL1           */
                                        /* P0.6=I2SRX_SDA     P0.7=I2STX_CLK      */
                                        /* P0.8=I2STX_WS                          */
    LPC_PINCON->PINSEL1 = 0x14000000;   /*                    P0.29=USB_D+        */
                                        /* P0.30=USB_D-                           */
    LPC_PINCON->PINSEL3 = 0x20000000;   /* P1.30=VBUS                             */
    LPC_PINCON->PINSEL4 = 0x00040000;   /*                    P2.9=USB_CONNECT    */
    LPC_PINCON->PINSEL9 = 0x04000000;   /*                    P4.29=TX_MCLK       */

    /*----- Internal pull-up and pull-down -------------------------------*/
    LPC_PINCON->PINMODE0 = 0x0002900A;  /* P0.0=---           P0.1=---            */
                                        /* P0.6=Repeater      P0.7=---            */
                                        /* P0.8=---                               */
    LPC_PINCON->PINMODE1 = 0x28002000;  /* P0.22=---                              */
                                        /*                    P0.29=---           */
                                        /* P0.30=---                              */
    LPC_PINCON->PINMODE2 = 0x002A0200;  /* P1.4=---                               */
                                        /* P1.8=---           P1.9=---            */
                                        /* P1.10=---                              */
    LPC_PINCON->PINMODE3 = 0x200000A0;  /* P1.18=---          P1.19=---           */
                                        /* P1.30=---                              */
    LPC_PINCON->PINMODE4 = 0x000AA800;  /*                    P2.5=---            */
                                        /* P2.6=---           P2.7=---            */
                                        /* P2.8=---           P2.9=---            */
    LPC_PINCON->PINMODE9 = 0x0A000000;  /* P4.28=---          P4.29=---           */

    /* ---- GPIO's -------------------------------------------------------*/
    LPC_GPIO0->FIODIR = (1 << 22) ;     /* P0.22=ENAREG1                          */
    LPC_GPIO0->FIOSET = (1 << 22);      /* Enable regulator1                      */
    LPC_GPIO1->FIODIR = 0x000C0710;     /* P1.4=IQSWAP                            */
                                        /* P1.8=TEILER2       P1.9=TEILER1        */
                                        /* P1.10=SI570_OE                         */
                                        /* P1.18=LED          P1.19=ENAREG2       */
    LPC_GPIO1->FIOSET = (1 << 19);      /* Enable regulator2                      */
    LPC_GPIO1->FIOSET = (1 << 10);      /* Enable Si570                           */
    LPC_GPIO2->FIODIR = 0x000001E0;     /*                    P2.5=PRESELECT1     */
                                        /* P2.6=PRESELECT2    P2.7=PRESELECT3     */
                                        /* P2.8=PRESELECT4                        */
    LPC_GPIO4->FIODIR &= ~0x10000000;   /* P4.28=UDA1361_PWON (floating = 0 dB)   */
#endif

    /*----- PLL's --------------------------------------------------------*/

    /* Select the flash wait states for desired CCLK */
    LPC_SC->FLASHCFG = (FLASH_WAITSTATES << 12) | 0x3A;

    /* Enable the crystal oscillator, and wait until it's running. */
    LPC_SC->SCS = 0x20;
    while( !(LPC_SC->SCS & (1 << 6)) )
    {
    }

    /* If the PLL is enabled or connected, disconnect and disable it now.
     * This can happen in debug mode.
     */
    if (LPC_SC->PLL0CON & (1 << 1)) {
        /* Disconnect now */
        LPC_SC->PLL0CON  = 1;
        LPC_SC->PLL0FEED = 0xAA;
        LPC_SC->PLL0FEED = 0x55;
    }
    /* Disable now */
    LPC_SC->PLL0CON  = 0;
    LPC_SC->PLL0FEED = 0xAA;
    LPC_SC->PLL0FEED = 0x55;

    /* Setup the Main PLL. Clock source is the crystal oscillator. */
    LPC_SC->CLKSRCSEL = 0x01;
    LPC_SC->PLL0CFG   = PLL0CFG_Val;
    LPC_SC->PLL0FEED  = 0xAA;
    LPC_SC->PLL0FEED  = 0x55;

    /* Clock diver to select CCLK */
    LPC_SC->CCLKCFG = (CLOCK_DIV - 1);

    /* Enable the Main PLL, and wait until it locks. Then connect it. */
    LPC_SC->PLL0CON   = 1;
    LPC_SC->PLL0FEED  = 0xAA;
    LPC_SC->PLL0FEED  = 0x55;
    while (!(LPC_SC->PLL0STAT & (1 << 26)))
    {
    }
    LPC_SC->PLL0CON   = 3;
    LPC_SC->PLL0FEED  = 0xAA;
    LPC_SC->PLL0FEED  = 0x55;

    /* Setup the USB PLL. Clock source is the crystal oscillator. */
    LPC_SC->PLL1CFG   = PLL1CFG_Val;
    LPC_SC->PLL1FEED  = 0xAA;
    LPC_SC->PLL1FEED  = 0x55;

    /* Enable the USB PLL, and wait until it locks. Then connect it. */
    LPC_SC->PLL1CON   = 1;
    LPC_SC->PLL1FEED  = 0xAA;
    LPC_SC->PLL1FEED  = 0x55;
    while (!(LPC_SC->PLL1STAT & (1 << 10)))
    {
    }
    LPC_SC->PLL1CON   = 3;
    LPC_SC->PLL1FEED  = 0xAA;
    LPC_SC->PLL1FEED  = 0x55;
}

