/*
  Softrock-40 Emulation
  Copyright (C) 2010  DF9DQ
  rolfm_9dq at yahoo dot de

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"

#include <stdio.h>
#include <string.h>
#include <stdbool.h>

#include "usbapi.h"
#include "usbdebug.h"
#include "usbstruct.h"
#include "usbhw_lpc.h"
#include "usbclass.h"
#include "i2c/si570.h"
#include "params.h"
#include "softrock.h"
#include "LPC17xx.h"


#define MAX_PACKET_SIZE	64
#include "../i2c/si570.h"


static unsigned char abVendorReqData[64];



/**
  Handle vendor requests

  @param [in] pSetup
  @param [out] piLen
  @param [out] ppbData
 */
void softrock_handleRequest (TSetupPacket *pSetup, unsigned char **ppbData)
{
    bool new_params;
    uint8_t *pbData = *ppbData;;
    uint16_t *pwData = (uint16_t *)pbData;
    uint32_t *pdData = (uint32_t *)pbData;
    int length;
    usb_Message message;
    uint32_t freq1121;
    int i;


    length = 0;
    new_params = false;

    if (REQTYPE_GET_RECIP(pSetup->bmRequestType) == REQTYPE_RECIP_DEVICE) {
        switch( pSetup->bRequest ) {
        case 0x00:  // Get firmware version
            pbData[0] = 12;
            pbData[1] = 15;
            length = 2;
        break;

        case 0x02:  /* Dummy */
            pbData[0] = 0x00;
            length = 1;
        break;

        case 0x15:
            pbData[0] = 0;
            pbData[1] = 0;
            length = 2;
        break;

        case 0x17:  /* Filter */
            /* First filter bank */
            if( pSetup->wIndex <= 255 )
            {
                /* Set a filter trip point? */
                if( pSetup->wIndex <= 2 )
                {
                    g_params.freqrxtrip[pSetup->wIndex] = pSetup->wValue;
                    new_params = true;
                }

                /* Always return current setting */
                pwData[ 0] = g_params.freqrxtrip[0];
                pwData[ 1] = g_params.freqrxtrip[1];
                pwData[ 2] = g_params.freqrxtrip[2];
                pwData[ 3] = 1;
                length = (3+1)*2;
            }
            /* Second filter bank (TX lowpass, not used) */
            else {
                pwData[ 0] = 0;
                length = 2;
            }
        break;

        case 0x18:  /* Set RX Band Pass Filter Address (ignored) */
            if (pSetup->wLength == 100) {
                for (i = 0; i < 100; i++) {
                    pbData[i] = 0;  /* No idea what the data should be... */
                }
                length = 100;
            }
            else {
                length = -2;    /* No idea how to handle this. --> Error */
            }
            break;
        
        case 0x19:  /* Read RX Band Pass Filter */
            pbData[0] = 0;
            pbData[1] = 1;
            pbData[2] = 2;
            pbData[3] = 3;
            length = 4;
        break;

        case 0x20:  /* Access to a (virtual) Si570 register */
            /* Response length should be 1 (flag that indicates error) */
            if (pSetup->wLength == 1) {
                si570_setSingleVirtualRegister(pSetup->wValue / 256, pSetup->wIndex % 256);
                pbData[0] = 0;      /* No error */
                length = 1;
            }
            else {
                length = -2;
            }
            break;
        
        case 0x30:  /* Set frequency by register values */
            si570_setVirtualRegisters (pbData);
            length = 0;
        break;

        case 0x31:  /* Set subtract/multiply values */
            freq1121 = si570_getFrequency();
            g_params.freqsubtract1121 = pdData[0];
            g_params.freqmultiply = pdData[1];
            si570_setFrequency(freq1121);
            length = 0;
            new_params = true;
        break;

        case 0x32:  /* Set oscillator frequency by value */
            si570_setFrequency(pdData[0]);
            length = 0;
        break;

        case 0x33:  /* Set XTAL frequency */
            g_params.freqxtal = pdData[0];
            si570_setFrequency(si570_getFrequency());   /* Update VCO with new XTAL frequency */
            length = 0;
            new_params = true;
        break;

        case 0x34:  /* Set startup frequency */
            g_params.freqstartup = pdData[0];
            length = 0;
            new_params = true;
        break;

        case 0x35:  /* Set smooth tune */
            g_params.smoothtune = pwData[0];
            length = 0;
            new_params = true;
        break;

        case 0x39:  /* Frequency subtract/multiply */
            pdData[0] = g_params.freqsubtract1121;
            pdData[1] = g_params.freqmultiply;
            length = 8;
        break;

        case 0x3A:  /* Read current frequency */
            pdData[0] = si570_getFrequency();
            length = 4;
        break;

        case 0x3B:  /* Read smooth tune */
            pwData[0] = g_params.smoothtune;
            length = 2;
        break;

        case 0x3C:  /* Read startup frequency */
            pdData[0] = g_params.freqstartup;
            length = 4;
        break;

        case 0x3D:  /* Read XTAL frequency */
            pdData[0] = g_params.freqxtal;
            length = 4;
        break;

        case 0x3F:  /* Read (virtual) Registers */
            si570_getVirtualRegisters();
            length = -1;
        break;

        case 0x41:  /* Set I2C address (ignored) */
            pbData[0] = si570_getAddress() >> 1;
            length = 1;
        break;

        case 0x42:  /* Dummy (Get CPU Temperature) */
            pwData[0] = 16 * 25;    /* 25°C */
            length = 2;
        break;

        case 0x50:  /* Set PTT line */
            pbData[0] = 0x22;
            length = 1;
        break;

        case 0x51:  /* Read CW key level */
            pbData[0] = 0xFF;
            length = 1;
        break;

        case 0xAB:    /* FiFi-SDR specific commands (read) */
            switch (pSetup->wIndex) {
            case 0:
                /* Return the SVN build number. */
                pdData[0] = SVNMAXREV;  /* This macro is defiuned in GCC command line (Makefile) */
                length = 4;
            break;
            case 1:
                /* Return a string describing the firmware version. */
                pbData[0] = 0;      /* TODO: Not yet supported */
                length = 1;
            break;
            case 3:
                pdData[0] = g_params.freq_3rd_harmonic;
                length = 4;
            break;
            case 5:
                pdData[0] = g_params.freq_5th_harmonic;
                length = 4;
            break;
            case 6:
                pdData[0] = g_params.presel_mode;
                length = 4;
            break;
            case 7:
                if (pSetup->wValue <= 15)
                {
                    pdData[0] = g_params.presel_freq[pSetup->wValue][0];
                    pdData[1] = g_params.presel_freq[pSetup->wValue][1];
                    pbData[8] = g_params.presel_pattern[pSetup->wValue];
                    length = 9;
                }
                else {
                    length = -2;
                }
            break;
            case 10:
                /* Read real Si570 registers */
                si570_getRegisters();
                length = -1;
            break;
            case 11:
                pdData[0] = g_params.virtual_vco_factor;
                length = 4;
            break;
            case 12:
                /* Read Si570 factory startup register values */
                si570_getFactoryDefaultRegisters();
                length = -1;
            break;
            case 15:
                /* Get demodulator mode (LSB/USB/...) */
                pbData[0] = usbclass_getDemodMode();
                length = 1;
            break;
            case 16:
                /* Get demodulator bandwidth (Hz) */
                pdData[0] = 2700ul;      // TODO
                length = 4;
            break;
            case 17:
                /* Get S-Meter */
                pdData[0] = usbclass_getRssi();      // TODO
                length = 4;
            break;
            case 18:
                /* Get FM center */
                pdData[0] = 42;      // TODO
                length = 4;
            break;
            case 19:
                /* Preamp (ADC 0/6 dB) */
                pbData[0] = 0;
                if (usbclass_getVolume() == 0x0600) {
                    pbData[0] = 1;
                }
                length = 1;
            break;
            case 0xFFFF:
                /* Debug data */
                pdData[0] = 0;
                pdData[1] = 0;
                pdData[2] = 0;
                pdData[3] = 0;
                pdData[4] = 0;
                pdData[5] = 0;
                pdData[6] = 0;
                pdData[7] = 0;
                length = 32;
            break;
            default:
                length = -2;
            break;
            }
        break;

        case 0xAC:    /* FiFi-SDR specific commands (write) */
            length = 0;
            switch (pSetup->wIndex) {
            case 3:
                g_params.freq_3rd_harmonic = pdData[0];
                new_params = true;
            break;
            case 5:
                g_params.freq_5th_harmonic = pdData[0];
                new_params = true;
            break;
            case 6:
                g_params.presel_mode = pdData[0];
                new_params = true;
            break;
            case 7:
                if (pSetup->wValue <= 15)
                {
                    g_params.presel_freq[pSetup->wValue][0] = pdData[0];
                    g_params.presel_freq[pSetup->wValue][1] = pdData[1];
                    g_params.presel_pattern[pSetup->wValue] = pbData[8];
                    if (pSetup->wValue == 15)
                    {
                        new_params = true;
                    }
                }
            break;
            case 11:
                g_params.virtual_vco_factor = pdData[0];
                new_params = true;
            break;
            case 13:
                /* Set IQ audio level (volume control) */
                message.opcode = USBMSG_SET_IQ_VOLUME;
                message.volume = pwData[0];
                xQueueSendToBack(usb_queue, &message, 0);
            break;
            case 14:
                /* Set demodulator audio level (volume control) */
                message.opcode = USBMSG_SET_DEMOD_VOLUME;
                message.volume = pwData[0];
                xQueueSendToBack(usb_queue, &message, 0);
            break;
            case 15:
                /* Set demodulator mode (LSB/USB/...) */
                message.opcode = USBMSG_SET_DEMOD_MODE;
                message.mode = pbData[0];
                xQueueSendToBack(usb_queue, &message, 0);
            break;
            case 16:
                /* Set demodulator bandwidth (Hz) */
                message.opcode = USBMSG_SET_DEMOD_BANDWIDTH;
                message.bandwidth = pdData[0];
                xQueueSendToBack(usb_queue, &message, 0);
            break;
            case 19:
                /* Preamp (ADC 0/6 dB) */
                message.opcode = USBMSG_SET_PREAMP;
                message.mode = pbData[0];
                xQueueSendToBack(usb_queue, &message, 0);
            break;
            default:
                length = -2;
            break;
            }
        break;
        default:
            length = -2;
        }
    }

    if (length == -2) {
        /* Error */
        StallControlPipe(0);
    }
    else if (length >= 0) {
        /* Limit length of response to maximum allowed by host. */
        if (pSetup->wLength < length) {
            length = pSetup->wLength;
        }

        usbcontrol_send_ep0(length, pbData);
    }

    /* Store new parameters */
    if (new_params) {
        si570_saveParams();
message.opcode = USBMSG_I2S_HOLD;
xQueueSendToBack(usb_queue, &message, 0);
    }

    ENABLE_EP_INT();
}



/**
  Handle vendor requests

  @param [in] pSetup
  @param [out] piLen
  @param [out] ppbData
 */
static bool SoftrockVendorRequest(TSetupPacket *pSetup, int *piLen, unsigned char **ppbData, portBASE_TYPE *taskWoken)
{
    usbclass_setupRequestFromISR(pSetup, ppbData, taskWoken);
    *piLen = -1;

    return true;
}



void softrock_init (void)
{
    /* Softrock communication uses vendor requests */
    USBRegisterRequestHandler (REQTYPE_TYPE_VENDOR, SoftrockVendorRequest, abVendorReqData);
}

